from flask import request, jsonify, g

from meltano.core.error import PluginInstallError
from meltano.core.plugin_discovery_service import (
    PluginDiscoveryService,
    PluginNotFoundError,
)
from meltano.core.plugin import PluginType
from meltano.core.project import Project
from meltano.core.project_add_service import ProjectAddService
from meltano.core.config_service import ConfigService
from meltano.core.plugin_install_service import (
    PluginInstallService,
    PluginInstallReason,
)
from flask_security import roles_required
from meltano.api.api_blueprint import APIBlueprint
from meltano.api.security.auth import block_if_readonly


pluginsBP = APIBlueprint("plugins", __name__)


@pluginsBP.errorhandler(PluginInstallError)
def _handle(ex):
    return (jsonify({"error": True, "code": str(ex)}), 502)


@pluginsBP.route("/all", methods=["GET"])
def all():
    project = Project.find()
    discovery = PluginDiscoveryService(project)

    def canonical_plugin(plugin):
        canonical_plugin = plugin.canonical()

        # let's remove all the settings related data
        canonical_plugin.pop("settings", None)
        canonical_plugin.pop("settings_group_validation", None)

        return canonical_plugin

    ordered_plugins = {
        plugin_type: [canonical_plugin(plugin) for plugin in plugins]
        for plugin_type, plugins in discovery.plugins_by_type().items()
    }

    return jsonify(ordered_plugins)


@pluginsBP.route("/installed", methods=["GET"])
def installed():
    """Returns JSON of all installed plugins

    Fuses the discovery.yml data with meltano.yml data and sorts each type alphabetically by name
    """

    project = Project.find()
    config = ConfigService(project)
    discovery = PluginDiscoveryService(project)

    def canonical_plugin(plugin):
        try:
            definition = discovery.find_plugin(plugin.type, plugin.name)
            canonical_plugin = {**definition.canonical(), **plugin.canonical()}
        except PluginNotFoundError:
            canonical_plugin = {**plugin.canonical()}

        canonical_plugin.pop("settings", None)
        canonical_plugin.pop("select", None)

        return canonical_plugin

    installed_plugins = {
        plugin_type: [
            canonical_plugin(plugin) for plugin in sorted(plugins, key=lambda x: x.name)
        ]
        for plugin_type, plugins in config.plugins_by_type().items()
    }

    return jsonify({**project.meltano.canonical(), "plugins": installed_plugins})


@pluginsBP.route("/add", methods=["POST"])
@block_if_readonly
def add():
    payload = request.get_json()
    plugin_type = PluginType(payload["plugin_type"])
    plugin_name = payload["name"]

    project = Project.find()
    add_service = ProjectAddService(project)
    plugin = add_service.add(plugin_type, plugin_name)

    return jsonify(plugin.canonical())


@pluginsBP.route("/install/batch", methods=["POST"])
@block_if_readonly
def install_batch():
    payload = request.get_json()
    plugin_type = PluginType(payload["plugin_type"])
    plugin_name = payload["name"]

    project = Project.find()

    config_service = ConfigService(project)
    plugin = config_service.find_plugin(plugin_name, plugin_type=plugin_type)

    add_service = ProjectAddService(project)
    related_plugins = add_service.add_related(plugin)

    # We will install the plugins in reverse order, since dependencies
    # are listed after their dependents in `related_plugins`, but should
    # be installed first.
    related_plugins.reverse()

    install_service = PluginInstallService(project)
    install_status = install_service.install_plugins(
        related_plugins, reason=PluginInstallReason.ADD
    )

    for error in install_status["errors"]:
        raise PluginInstallError(error["message"])

    return jsonify([plugin.canonical() for plugin in related_plugins])


@pluginsBP.route("/install", methods=["POST"])
@block_if_readonly
def install():
    payload = request.get_json()
    plugin_type = PluginType(payload["plugin_type"])
    plugin_name = payload["name"]

    project = Project.find()

    config_service = ConfigService(project)
    plugin = config_service.find_plugin(plugin_name, plugin_type=plugin_type)

    install_service = PluginInstallService(project)
    install_service.install_plugin(plugin, reason=PluginInstallReason.ADD)

    return jsonify(plugin.canonical())



import asyncio
import click
import datetime
import logging
import os
import sys
from contextlib import suppress, contextmanager
from async_generator import asynccontextmanager

from . import cli
from .utils import CliError, add_plugin, add_related_plugins, install_plugins
from .params import project
from meltano.core.config_service import ConfigService
from meltano.core.runner import RunnerError
from meltano.core.runner.singer import SingerRunner
from meltano.core.runner.dbt import DbtRunner
from meltano.core.project import Project, ProjectNotFound
from meltano.core.job import Job
from meltano.core.plugin import PluginRef, PluginType
from meltano.core.plugin.error import PluginMissingError
from meltano.core.project_add_service import ProjectAddService
from meltano.core.transform_add_service import TransformAddService
from meltano.core.tracking import GoogleAnalyticsTracker
from meltano.core.db import project_engine
from meltano.core.plugin_discovery_service import (
    PluginDiscoveryService,
    PluginNotFoundError,
)
from meltano.core.plugin_install_service import PluginInstallReason
from meltano.core.logging import OutputLogger, JobLoggingService
from meltano.core.elt_context import ELTContextBuilder


logger = logging.getLogger(__name__)


def logs(*args, **kwargs):
    logger.info(click.style(*args, **kwargs))


@cli.command()
@click.argument("extractor")
@click.argument("loader")
@click.option("--transform", type=click.Choice(["skip", "only", "run"]), default="skip")
@click.option("--dry", help="Do not actually run.", is_flag=True)
@click.option(
    "--full-refresh",
    help="Perform a full refresh (ignore state left behind by any previous runs)",
    is_flag=True,
)
@click.option(
    "--select",
    "-s",
    help="Select only these specific entities for extraction",
    multiple=True,
    default=[],
)
@click.option(
    "--exclude",
    "-e",
    help="Exclude these specific entities from extraction",
    multiple=True,
    default=[],
)
@click.option(
    "--job_id", envvar="MELTANO_JOB_ID", help="A custom string to identify the job."
)
@project(migrate=True)
def elt(
    project, extractor, loader, transform, dry, full_refresh, select, exclude, job_id
):
    """
    meltano elt EXTRACTOR_NAME LOADER_NAME

    extractor_name: Which extractor should be used in this extraction
    loader_name: Which loader should be used in this extraction
    """

    select_filter = [*select, *(f"!{entity}" for entity in exclude)]

    job = Job(
        job_id=job_id
        or f'{datetime.datetime.utcnow().strftime("%Y-%m-%dT%H%M%S")}--{extractor}--{loader}'
    )

    _, Session = project_engine(project)
    session = Session()

    try:
        job_logging_service = JobLoggingService(project)

        with job.run(session), job_logging_service.create_log(
            job.job_id, job.run_id
        ) as log_file:
            output_logger = OutputLogger(log_file)

            run_async(
                run_elt(
                    project,
                    job,
                    extractor,
                    loader,
                    transform,
                    output_logger=output_logger,
                    session=session,
                    dry_run=dry,
                    full_refresh=full_refresh,
                    select_filter=select_filter,
                )
            )
    finally:
        session.close()

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_elt(extractor=extractor, loader=loader, transform=transform)


def run_async(coro):
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(coro)
    except asyncio.CancelledError:
        pass
    finally:
        # The below is taken from https://stackoverflow.com/a/58532304
        # and inspired by to Python 3.7's `asyncio.run`
        all_tasks = asyncio.gather(
            *asyncio.Task.all_tasks(loop), return_exceptions=True
        )
        all_tasks.cancel()
        with suppress(asyncio.CancelledError):
            loop.run_until_complete(all_tasks)
        loop.run_until_complete(loop.shutdown_asyncgens())


@asynccontextmanager
async def redirect_output(output_logger):
    meltano_stdout = output_logger.out("meltano", stream=sys.stdout, color="blue")
    meltano_stderr = output_logger.out("meltano", color="blue")

    with meltano_stdout.redirect_logging(ignore_errors=(CliError,)):
        async with meltano_stdout.redirect_stdout(), meltano_stderr.redirect_stderr():
            yield


async def run_elt(
    project,
    job,
    extractor,
    loader,
    transform,
    output_logger,
    session,
    dry_run=False,
    full_refresh=False,
    select_filter=[],
):
    config_service = ConfigService(project)
    discovery_service = PluginDiscoveryService(project, config_service=config_service)

    plugin_refs = []
    if transform != "only":
        plugin_refs.append(PluginRef(PluginType.EXTRACTORS, extractor))
        plugin_refs.append(PluginRef(PluginType.LOADERS, loader))

    transform_name = None
    if transform != "skip":
        transform_name = find_transform_for_extractor(
            extractor, discovery_service=discovery_service
        )

        if transform_name:
            plugin_refs.append(PluginRef(PluginType.TRANSFORMS, transform_name))
        else:
            # There is no default transform for this extractor..
            # Don't panic, everything is cool - just run custom transforms
            plugin_refs.append(PluginRef(PluginType.TRANSFORMERS, "dbt"))

    async with redirect_output(output_logger):
        try:
            success = install_missing_plugins(
                project,
                plugin_refs,
                config_service=config_service,
                discovery_service=discovery_service,
            )

            if not success:
                raise CliError("Failed to install missing plugins")

            elt_context = (
                ELTContextBuilder(project)
                .with_job(job)
                .with_extractor(extractor)
                .with_loader(loader)
                .with_transform(transform_name or transform)
                .with_dry_run(dry_run)
                .with_full_refresh(full_refresh)
                .with_select_filter(select_filter)
                .context(session)
            )

            if transform != "only":
                await run_extract_load(elt_context, output_logger, session)
            else:
                logs("Extract & load skipped.", fg="yellow")

            if elt_context.transformer:
                await run_transform(elt_context, output_logger, session)
            else:
                logs("Transformation skipped.", fg="yellow")
        except RunnerError as err:
            raise CliError(f"ELT could not be completed: {err}") from err


async def run_extract_load(elt_context, output_logger, session, **kwargs):
    extractor = elt_context.extractor.name
    loader = elt_context.loader.name

    extractor_log = output_logger.out(extractor, color="yellow")
    loader_log = output_logger.out(loader, color="green")

    @contextmanager
    def nullcontext():
        yield None

    extractor_out_writer = nullcontext
    loader_out_writer = nullcontext
    if logger.getEffectiveLevel() == logging.DEBUG:
        extractor_out = output_logger.out(f"{extractor} (out)", color="bright_yellow")
        loader_out = output_logger.out(f"{loader} (out)", color="bright_green")

        extractor_out_writer = extractor_out.line_writer
        loader_out_writer = loader_out.line_writer

    logs("Running extract & load...")

    singer_runner = SingerRunner(elt_context)
    try:
        with extractor_log.line_writer() as extractor_log_writer, loader_log.line_writer() as loader_log_writer:
            with extractor_out_writer() as extractor_out_writer, loader_out_writer() as loader_out_writer:
                await singer_runner.run(
                    session,
                    **kwargs,
                    extractor_log=extractor_log_writer,
                    loader_log=loader_log_writer,
                    extractor_out=extractor_out_writer,
                    loader_out=loader_out_writer,
                )
    except RunnerError as err:
        try:
            code = err.exitcodes["extractor"]
            message = extractor_log.last_line.rstrip() or "(see above)"
            logger.error(
                f"{click.style(f'Extraction failed ({code}):', fg='red')} {message}"
            )
        except KeyError:
            pass

        try:
            code = err.exitcodes["loader"]
            message = loader_log.last_line.rstrip() or "(see above)"
            logger.error(
                f"{click.style(f'Loading failed ({code}):', fg='red')} {message}"
            )
        except KeyError:
            pass

        raise

    logs("Extract & load complete!", fg="green")


async def run_transform(elt_context, output_logger, session, **kwargs):
    transformer_log = output_logger.out(elt_context.transformer.name, color="magenta")

    logs("Running transformation...")

    dbt_runner = DbtRunner(elt_context)
    try:
        with transformer_log.line_writer() as transformer_log_writer:
            await dbt_runner.run(session, **kwargs, log=transformer_log_writer)
    except RunnerError as err:
        try:
            code = err.exitcodes["transformer"]
            message = transformer_log.last_line.rstrip() or "(see above)"
            logger.error(
                f"{click.style(f'Transformation failed ({code}):', fg='red')} {message}"
            )
        except KeyError:
            pass

        raise

    logs("Transformation complete!", fg="green")


def find_transform_for_extractor(extractor: str, discovery_service):
    try:
        extractor_plugin_def = discovery_service.find_plugin(
            PluginType.EXTRACTORS, extractor
        )
        # Check if there is a default transform for this extractor
        transform_plugin_def = discovery_service.find_plugin_by_namespace(
            PluginType.TRANSFORMS, extractor_plugin_def.namespace
        )

        return transform_plugin_def.name
    except PluginNotFoundError:
        return None


def add_plugin_if_missing(
    project: Project,
    plugin_type,
    name,
    config_service: ConfigService,
    add_service: ProjectAddService,
):
    try:
        config_service.find_plugin(name, plugin_type=plugin_type)
        return None
    except PluginMissingError:
        logs(
            f"{plugin_type.descriptor.capitalize()} '{name}' is missing, adding it to your project...",
            fg="yellow",
        )
        return add_plugin(project, plugin_type, name, add_service=add_service)


def install_missing_plugins(
    project: Project,
    plugin_refs: [PluginRef],
    config_service: ConfigService,
    discovery_service: PluginDiscoveryService,
):
    add_service = ProjectAddService(
        project,
        plugin_discovery_service=discovery_service,
        config_service=config_service,
    )

    plugins = []
    for plugin_ref in plugin_refs:
        plugin = add_plugin_if_missing(
            project,
            plugin_ref.type,
            plugin_ref.name,
            config_service=config_service,
            add_service=add_service,
        )
        if plugin:
            plugins.append(plugin)

    if not plugins:
        return True

    related_plugins = add_related_plugins(
        project, plugins, add_service=add_service, plugin_types=[PluginType.FILES]
    )
    plugins.extend(related_plugins)

    # We will install the plugins in reverse order, since dependencies
    # are listed after their dependents in `related_plugins`, but should
    # be installed first.
    plugins.reverse()

    return install_plugins(project, plugins, reason=PluginInstallReason.ADD)



import os
import json
import yaml
import logging
from typing import List

from .project import Project
from .plugin import PluginType, Plugin, PluginInstall, PluginRef
from .plugin_discovery_service import PluginDiscoveryService
from .plugin.factory import plugin_factory
from .config_service import ConfigService, PluginAlreadyAddedException


class PluginNotSupportedException(Exception):
    pass


class MissingPluginException(Exception):
    pass


class ProjectAddService:
    def __init__(
        self,
        project: Project,
        plugin_discovery_service: PluginDiscoveryService = None,
        config_service: ConfigService = None,
    ):
        self.project = project
        self.discovery_service = plugin_discovery_service or PluginDiscoveryService(
            project
        )
        self.config_service = config_service or ConfigService(project)

    def add(self, plugin_type: PluginType, plugin_name: str, **kwargs) -> PluginInstall:
        plugin = self.discovery_service.find_plugin(plugin_type, plugin_name)
        installed = plugin.as_installed()
        return self.config_service.add_to_file(installed)

    def add_custom(
        self, plugin_type: PluginType, plugin_name: str, namespace: str, **kwargs
    ) -> PluginInstall:
        plugin = Plugin(plugin_type, plugin_name, namespace, **kwargs)
        installed = plugin.as_installed(custom=True)
        return self.config_service.add_to_file(installed)

    def add_related(
        self,
        target_plugin: PluginInstall,
        plugin_types: List[PluginType] = list(PluginType),
    ):
        try:
            plugin_types.remove(target_plugin.type)
        except ValueError:
            pass

        related_plugins = []

        runner = target_plugin.runner
        if runner:
            related_plugins.append(runner)

        plugin_def = self.discovery_service.find_plugin(
            target_plugin.type, target_plugin.name
        )
        related_plugins.extend(
            plugin
            for plugin_type in plugin_types
            for plugin in self.discovery_service.get_plugins_of_type(plugin_type)
            if plugin.namespace == plugin_def.namespace
        )

        added_plugins = []
        for plugin in related_plugins:
            try:
                plugin_install = self.add(plugin.type, plugin.name)
            except PluginAlreadyAddedException as err:
                continue

            added_plugins.append(plugin_install)

        added_plugins_with_related = []
        for plugin_install in added_plugins:
            added_plugins_with_related.extend(
                [
                    plugin_install,
                    *self.add_related(plugin_install, plugin_types=plugin_types),
                ]
            )

        return added_plugins_with_related



import json
import logging
import subprocess
from typing import Dict
from jsonschema import Draft4Validator

from meltano.core.setting_definition import SettingDefinition
from meltano.core.behavior.hookable import hook
from meltano.core.plugin.error import PluginExecutionError, PluginLacksCapabilityError
from meltano.core.plugin_invoker import InvokerError
from meltano.core.utils import file_has_data, truthy, flatten

from . import SingerPlugin, PluginType
from .catalog import (
    MetadataExecutor,
    MetadataRule,
    SchemaExecutor,
    SchemaRule,
    property_breadcrumb,
    select_metadata_rules,
)

logger = logging.getLogger(__name__)


def config_metadata_rules(config):
    flat_config = flatten(config, "dot")

    rules = []
    for key, value in flat_config.items():
        # <tap_stream_id>.<key>
        # <tap_stream_id>.<prop>.<key>
        # <tap_stream_id>.<prop>.<subprop>.<key>
        # <tap_stream_id>.properties.<prop>.<key>
        # <tap_stream_id>.properties.<prop>.properties.<subprop>.<key>
        tap_stream_id, *props, key = key.split(".")

        rules.append(
            MetadataRule(
                tap_stream_id=tap_stream_id,
                breadcrumb=property_breadcrumb(props),
                key=key,
                value=value,
            )
        )

    return rules


def config_schema_rules(config):
    return [
        SchemaRule(
            tap_stream_id=tap_stream_id,
            breadcrumb=["properties", prop],
            payload=payload,
        )
        for tap_stream_id, stream_config in config.items()
        for prop, payload in stream_config.items()
    ]


def select_filter_metadata_rules(tap_stream_ids):
    if not tap_stream_ids:
        return []

    metadata_rules = select_metadata_rules(
        [f"{tap_stream_id}." for tap_stream_id in tap_stream_ids]
    )

    has_include_rules = any(rule.value for rule in metadata_rules)
    if has_include_rules:
        metadata_rules = [
            # First deselect all stream nodes, without touching their property nodes
            *select_metadata_rules(["!*."]),
            # Then (de)select only the specified stream nodes
            *metadata_rules,
        ]

    return metadata_rules


class SingerTap(SingerPlugin):
    __plugin_type__ = PluginType.EXTRACTORS

    @property
    def extra_settings(self):
        return [
            SettingDefinition(
                name="_load_schema", value="$MELTANO_EXTRACTOR_NAMESPACE"
            ),
            SettingDefinition(name="_select", kind="array", value=["*.*"]),
            SettingDefinition(
                name="_metadata",
                aliases=["metadata"],
                kind="object",
                value={},
                value_processor="nest_object",
            ),
            SettingDefinition(
                name="_schema", kind="object", value={}, value_processor="nest_object"
            ),
            SettingDefinition(name="_select_filter", kind="array", value=[]),
        ]

    def exec_args(self, plugin_invoker):
        """
        Return the arguments list with the complete runtime paths.
        """
        args = ["--config", plugin_invoker.files["config"]]

        catalog_path = plugin_invoker.files["catalog"]
        if file_has_data(catalog_path):
            if "catalog" in plugin_invoker.capabilities:
                args += ["--catalog", catalog_path]
            elif "properties" in plugin_invoker.capabilities:
                args += ["--properties", catalog_path]
            else:
                logger.warn(
                    "A catalog file was found, but it will be ignored as the extractor does not advertise the `catalog` or `properties` capability"
                )

        state_path = plugin_invoker.files["state"]
        if file_has_data(state_path):
            if "state" in plugin_invoker.capabilities:
                args += ["--state", state_path]
            else:
                logger.warn(
                    "A state file was found, but it will be ignored as the extractor does not advertise the `state` capability"
                )

        return args

    @property
    def config_files(self):
        return {
            "config": "tap.config.json",
            "catalog": "tap.properties.json",
            "state": "state.json",
        }

    @property
    def output_files(self):
        return {"output": "tap.out"}

    @hook("before_invoke", can_fail=PluginExecutionError)
    def run_discovery_hook(self, plugin_invoker, exec_args=[]):
        if "--discover" in exec_args or "--help" in exec_args:
            return

        try:
            self.run_discovery(plugin_invoker, exec_args)
        except PluginLacksCapabilityError:
            pass

    def run_discovery(self, plugin_invoker, exec_args=[]):
        if not "discover" in plugin_invoker.capabilities:
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support catalog discovery (the `discover` capability is not advertised)"
            )

        properties_file = plugin_invoker.files["catalog"]

        try:
            with properties_file.open("w") as catalog:
                result = plugin_invoker.invoke(
                    "--discover",
                    stdout=catalog,
                    stderr=subprocess.PIPE,
                    universal_newlines=True,
                )
                stdout, stderr = result.communicate()
                exit_code = result.returncode
        except Exception:
            properties_file.unlink()
            raise

        if exit_code != 0:
            properties_file.unlink()
            raise PluginExecutionError(
                f"Catalog discovery failed: command {plugin_invoker.exec_args('--discover')} returned {exit_code}: {stderr.rstrip()}"
            )

        # test for the schema to be a valid catalog
        try:
            with properties_file.open("r") as catalog:
                catalog_json = json.load(catalog)
                schema_valid = Draft4Validator.check_schema(catalog_json)
        except Exception as err:
            properties_file.unlink()
            raise PluginExecutionError(
                f"Catalog discovery failed: invalid catalog output by --discover: {err}"
            ) from err

    @hook("before_invoke", can_fail=PluginExecutionError)
    def apply_catalog_rules_hook(self, plugin_invoker, exec_args=[]):
        if "--discover" in exec_args or "--help" in exec_args:
            return

        try:
            self.apply_catalog_rules(plugin_invoker, exec_args)
        except PluginLacksCapabilityError:
            pass

    def apply_catalog_rules(self, plugin_invoker, exec_args=[]):
        if (
            not "catalog" in plugin_invoker.capabilities
            and not "properties" in plugin_invoker.capabilities
        ):
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support entity selection or catalog metadata and schema rules"
            )

        properties_file = plugin_invoker.files["catalog"]

        try:
            with properties_file.open() as catalog:
                schema = json.load(catalog)

            config = plugin_invoker.plugin_config_extras

            schema_rules = config_schema_rules(config["_schema"])
            SchemaExecutor(schema_rules).visit(schema)

            metadata_rules = [
                *select_metadata_rules(["!*.*"]),
                *select_metadata_rules(config["_select"]),
                *config_metadata_rules(config["_metadata"]),
                *select_filter_metadata_rules(config["_select_filter"]),
            ]
            MetadataExecutor(metadata_rules).visit(schema)

            with properties_file.open("w") as catalog:
                json.dump(schema, catalog, indent=2)
        except FileNotFoundError as err:
            raise PluginExecutionError(
                f"Applying catalog rules failed: catalog file is missing."
            ) from err
        except Exception as err:
            properties_file.unlink()
            raise PluginExecutionError(
                f"Applying catalog rules failed: catalog file is invalid: {err}"
            ) from err
